@@ -6,7 +6,9 @@ from redmine.exceptions import (
     ResourceAttrError,
     ReadonlyAttrError,
     CustomFieldValueError,
-    ResourceVersionMismatchError
+    ResourceVersionMismatchError,
+    ResourceNotFoundError,
+    ResourceRequirementsError
 )
 
 # Resources which when accessed from some other
@@ -24,6 +26,12 @@ _RESOURCE_SET_MAP = {
     'journals': 'IssueJournal',
     'children': 'Issue',
     'roles': 'Role',
+    'issues': 'Issue',
+    'projects': 'Project',
+    'notes': 'Note',
+    'deals': 'Deal',
+    'contacts': 'Contact',
+    'related_contacts': 'Contact',
 }
 
 # Resources which when accessed from some other
@@ -40,9 +48,10 @@ _RESOURCE_MAP = {
     'activity': 'Enumeration',
     'category': 'IssueCategory',
     'fixed_version': 'Version',
+    'contact': 'Contact',
 }
 
-# Resources which when access from some other
+# Resources which when accessed from some other
 # resource should be requested from Redmine
 _RESOURCE_RELATIONS_MAP = {
     'wiki_pages': 'WikiPage',
@@ -53,6 +62,9 @@ _RESOURCE_RELATIONS_MAP = {
     'relations': 'IssueRelation',
     'time_entries': 'TimeEntry',
     'issues': 'Issue',
+    'contacts': 'Contact',
+    'deals': 'Deal',
+    'deal_categories': 'DealCategory',
 }
 
 # Resource attributes which when set should
@@ -68,6 +80,8 @@ _RESOURCE_SINGLE_ATTR_ID_MAP = {
     'parent_issue_id': 'parent',
     'issue_id': 'issue',
     'activity_id': 'activity',
+    'status_id': 'status',
+    'contact_id': 'contact',
 }
 
 # Resource attributes which when set should
@@ -96,6 +110,7 @@ class _Resource(object):
 
     _includes = ()
     _relations = ()
+    _relations_name = None
     _unconvertible = ()
     _members = ('manager',)
     _create_readonly = ('id', 'created_on', 'updated_on', 'author', 'user', 'project', 'issue')
@@ -112,12 +127,15 @@ class _Resource(object):
         self._update_readonly += self._relations + self._includes
         self._changes = {}
 
+        if self._relations_name is None:
+            self._relations_name = self.__class__.__name__.lower()
+
     def __getitem__(self, item):
-        """Provides a dictionary like access to resource attributes"""
+        """Provides a dictionary-like access to resource attributes"""
         return getattr(self, item)
 
     def __setitem__(self, item, value):
-        """Provides a dictionary like setter for resource attributes"""
+        """Provides a dictionary-like setter for resource attributes"""
         return setattr(self, item, value)
 
     def __getattr__(self, item):
@@ -139,7 +157,7 @@ class _Resource(object):
 
             # If item is a relation and should be requested from Redmine, let's do it
             elif item in self._relations and self._attributes[item] is None:
-                filters = {'{0}_id'.format(self.__class__.__name__.lower()): self.internal_id}
+                filters = {'{0}_id'.format(self._relations_name): self.internal_id}
                 manager = ResourceManager(self.manager.redmine, _RESOURCE_RELATIONS_MAP[item])
                 self._attributes[item] = manager.filter(**filters)
                 return self._attributes[item]
@@ -312,7 +330,18 @@ class Project(_Resource):
     query_delete = '/projects/{0}.json'
 
     _includes = ('trackers', 'issue_categories')
-    _relations = ('wiki_pages', 'memberships', 'issue_categories', 'versions', 'news', 'issues')
+    _relations = (
+        'wiki_pages',
+        'memberships',
+        'issue_categories',
+        'time_entries',
+        'versions',
+        'news',
+        'issues',
+        'contacts',
+        'deals',
+        'deal_categories',
+    )
     _unconvertible = ('status',)
     _update_readonly = _Resource._update_readonly + ('identifier',)
 
@@ -339,6 +368,7 @@ class Issue(_Resource):
 
     _includes = ('children', 'attachments', 'relations', 'changesets', 'journals', 'watchers')
     _relations = ('relations', 'time_entries')
+    _unconvertible = ('notes',)
     _create_readonly = _Resource._create_readonly + ('spent_hours',)
     _update_readonly = _create_readonly
 
@@ -361,6 +391,13 @@ class Issue(_Resource):
             url = '{0}/issues/{1}/watchers/{2}.json'.format(self._redmine.url, self._issue_id, user_id)
             return self._redmine.request('delete', url)
 
+    @classmethod
+    def translate_params(cls, params):
+        if 'version_id' in params:
+            params['fixed_version_id'] = params.pop('version_id')
+
+        return super(Issue, cls).translate_params(params)
+
     def __getattr__(self, item):
         if item == 'version':
             return super(Issue, self).__getattr__('fixed_version')
@@ -639,10 +676,21 @@ class User(_Resource):
     query_delete = '/users/{0}.json'
 
     _includes = ('memberships', 'groups')
+    _relations = ('issues', 'time_entries', 'contacts', 'deals')
+    _relations_name = 'assigned_to'
     _unconvertible = ('status',)
     _create_readonly = _Resource._create_readonly + ('api_key', 'last_login_on')
     _update_readonly = _create_readonly
 
+    def __getattr__(self, item):
+        if item == 'time_entries':
+            self._relations_name = 'user'
+            value = super(User, self).__getattr__(item)
+            self._relations_name = 'assigned_to'
+            return value
+
+        return super(User, self).__getattr__(item)
+
     def __str__(self):
         try:
             return super(User, self).__str__()
@@ -732,6 +780,9 @@ class IssueStatus(_Resource):
     container_all = 'issue_statuses'
     query_all = '/issue_statuses.json'
 
+    _relations = ('issues',)
+    _relations_name = 'status'
+
     @property
     def url(self):
         return '{0}/issue_statuses/{1}/edit'.format(self.manager.redmine.url, self.internal_id)
@@ -742,6 +793,8 @@ class Tracker(_Resource):
     container_all = 'trackers'
     query_all = '/trackers.json'
 
+    _relations = ('issues',)
+
     @property
     def url(self):
         return '{0}/trackers/{1}/edit'.format(self.manager.redmine.url, self.internal_id)
@@ -775,7 +828,234 @@ class CustomField(_Resource):
         # i.e. project, and it's not used in the resource, there will be
         # no value attribute defined, that is why we need to return 0 or
         # we'll get an exception
-        if item == 'value' and not item in self._attributes:
+        if item == 'value' and item not in self._attributes:
             return 0
 
         return super(CustomField, self).__getattr__(item)
+
+
+class Note(_Resource):
+    redmine_version = '2.1'
+    requirements = (('CRM plugin', '3.2.4'),)
+    container_one = 'note'
+    query_one = '/notes/{0}.json'
+
+    def __getattr__(self, item):
+        if item == 'source' and item in self._attributes and self._attributes[item].get('type') in ('Deal', 'Contact'):
+            manager = ResourceManager(self.manager.redmine, self._attributes[item]['type'])
+            return manager.to_resource(self._attributes[item])
+
+        return super(Note, self).__getattr__(item)
+
+    def __str__(self):
+        return self.content
+
+    def __repr__(self):
+        return '<{0}.{1} #{2}>'.format(
+            self.__class__.__module__,
+            self.__class__.__name__,
+            self.id
+        )
+
+
+class Contact(_Resource):
+    redmine_version = '1.2.1'
+    requirements = ('CRM plugin',)
+    container_all = 'contacts'
+    container_one = 'contact'
+    container_filter = 'contacts'
+    container_create = 'contact'
+    container_update = 'contact'
+    query_all = '/contacts.json'
+    query_one = '/contacts/{0}.json'
+    query_filter = '/contacts.json'
+    query_create = '/projects/{project_id}/contacts.json'
+    query_update = '/contacts/{0}.json'
+    query_delete = '/contacts/{0}.json'
+
+    _includes = ('notes', 'contacts', 'deals', 'issues')
+
+    class Project:
+        """A contact project implementation"""
+        def __init__(self, contact):
+            self._redmine = contact.manager.redmine
+            self._contact_id = contact.internal_id
+
+            if self._redmine.ver is not None and LooseVersion(str(self._redmine.ver)) < LooseVersion('2.3'):
+                raise ResourceVersionMismatchError
+
+        def add(self, project_id):
+            """Adds project to contact's project list"""
+            url = '{0}/contacts/{1}/projects.json'.format(self._redmine.url, self._contact_id)
+
+            try:
+                return self._redmine.request('post', url, data={'project': {'id': project_id}})
+            except ResourceNotFoundError:
+                raise ResourceRequirementsError((('CRM plugin', '3.4.0'),))
+
+        def remove(self, project_id):
+            """Removes project from contact's project list"""
+            url = '{0}/contacts/{1}/projects/{2}.json'.format(self._redmine.url, self._contact_id, project_id)
+
+            try:
+                return self._redmine.request('delete', url)
+            except ResourceNotFoundError:
+                raise ResourceRequirementsError((('CRM plugin', '3.4.0'),))
+
+    @classmethod
+    def translate_params(cls, params):
+        if 'tag_list' in params:
+            params['tag_list'] = ','.join(params['tag_list'])
+
+        if 'phones' in params:
+            params['phone'] = ','.join(params.pop('phones'))
+
+        if 'emails' in params:
+            params['email'] = ','.join(params.pop('emails'))
+
+        return super(Contact, cls).translate_params(params)
+
+    def __getattr__(self, item):
+        if item == 'project':
+            return Contact.Project(self)
+        elif item == 'phones':
+            return [p.get('number') if isinstance(p, dict) else p for p in self._attributes.get('phones', [])]
+        elif item == 'emails':
+            return [e.get('address') if isinstance(e, dict) else e for e in self._attributes.get('emails', [])]
+        elif item == 'avatar' and item in self._attributes:
+            manager = ResourceManager(self.manager.redmine, 'Attachment')
+            return manager.to_resource({'id': self._attributes[item].get('attachment_id', 0)})
+
+        return super(Contact, self).__getattr__(item)
+
+    def __setattr__(self, item, value):
+        if item in ('phones', 'emails', 'tag_list'):
+            self._attributes[item] = value
+            self._changes.update(self.translate_params({item: value}))
+        else:
+            super(Contact, self).__setattr__(item, value)
+
+    def __str__(self):
+        try:
+            return super(Contact, self).__str__()
+        except ResourceAttrError:
+            if not getattr(self, 'last_name', False):
+                return '{0}'.format(to_string(self.first_name))
+            else:
+                return '{0} {1}'.format(to_string(self.first_name), to_string(self.last_name))
+
+    def __repr__(self):
+        try:
+            return super(Contact, self).__repr__()
+        except ResourceAttrError:
+            if not getattr(self, 'last_name', False):
+                return '<{0}.{1} #{2} "{3}">'.format(
+                    self.__class__.__module__,
+                    self.__class__.__name__,
+                    self.id,
+                    to_string(self.first_name),
+                )
+            else:
+                return '<{0}.{1} #{2} "{3} {4}">'.format(
+                    self.__class__.__module__,
+                    self.__class__.__name__,
+                    self.id,
+                    to_string(self.first_name),
+                    to_string(self.last_name)
+                )
+
+
+class ContactTag(_Resource):
+    redmine_version = '2.3'
+    requirements = (('CRM plugin', '3.4.0'),)
+    container_all = 'tags'
+    query_all = '/contacts_tags.json'
+
+    @property
+    def url(self):
+        return '{0}/contacts_tags/{1}/edit'.format(self.manager.redmine.url, self.internal_id)
+
+
+class CrmQuery(_Resource):
+    redmine_version = '2.3'
+    requirements = (('CRM plugin', '3.3.0'),)
+    container_filter = 'queries'
+    query_filter = '/crm_queries.json?object_type={resource}'
+
+    _relations = ('deals',)
+    _relations_name = 'query'
+
+    @property
+    def url(self):
+        return '{0}/projects/{1}/{2}s?query_id={3}'.format(
+            self.manager.redmine.url,
+            self._attributes.get('project_id', 0),
+            self.manager.params.get('resource', ''),
+            self.internal_id
+        )
+
+
+class Deal(_Resource):
+    redmine_version = '1.2.1'
+    requirements = ('CRM plugin',)
+    container_all = 'deals'
+    container_one = 'deal'
+    container_filter = 'deals'
+    container_create = 'deal'
+    container_update = 'deal'
+    query_all = '/deals.json'
+    query_one = '/deals/{0}.json'
+    query_filter = '/deals.json'
+    query_create = '/projects/{project_id}/deals.json'
+    query_update = '/deals/{0}.json'
+    query_delete = '/deals/{0}.json'
+
+    _includes = ('notes',)
+
+    def __getattr__(self, item):
+        if item in ('category', 'status') and item in self._attributes:
+            manager = ResourceManager(self.manager.redmine, 'Deal{0}'.format(item.capitalize()))
+            return manager.to_resource(self._attributes[item])
+
+        return super(Deal, self).__getattr__(item)
+
+    def __str__(self):
+        try:
+            return super(Deal, self).__str__()
+        except ResourceAttrError:
+            return str(self.id)
+
+    def __repr__(self):
+        try:
+            return super(Deal, self).__repr__()
+        except ResourceAttrError:
+            return '<{0}.{1} #{2}>'.format(
+                self.__class__.__module__,
+                self.__class__.__name__,
+                self.id
+            )
+
+
+class DealStatus(_Resource):
+    redmine_version = '2.3'
+    requirements = (('CRM plugin', '3.3.0'),)
+    container_all = 'deal_statuses'
+    query_all = '/deal_statuses.json'
+
+    _relations = ('deals',)
+    _relations_name = 'status'
+
+    @property
+    def url(self):
+        return '{0}/deal_statuses/{1}/edit'.format(self.manager.redmine.url, self.internal_id)
+
+
+class DealCategory(_Resource):
+    redmine_version = '2.3'
+    requirements = (('CRM plugin', '3.3.0'),)
+    container_filter = 'deal_categories'
+    query_filter = '/projects/{project_id}/deal_categories.json'
+
+    @property
+    def url(self):
+        return '{0}/deal_categories/edit?id={1}'.format(self.manager.redmine.url, self.internal_id)
